/* simple minded xargs(1)                                          *
 * Copyright by Uwe Reder (uereder@cip.informatik.uni-erlangen.de) *
 * xargs [-tp] [-ireplstr] [command [arg...]]                       *
 *                                                                 *
 * Permission is granted to make and distribute verbatim copies of *
 * this provided the copyright notice is preserved on all copies.  */

#include<stdio.h>
#include<string.h>
#include<types.h>
#include<wait.h>

#define DEFAULT_REPLSTR    "{}"
#define DEFAULT_SEPERATORS " \t\n"
#define DEFAULT_TRACE      0
#define DEFAULT_INSERT     0
#define DEFAULT_PROMPT     0

#define ECHO_TO stderr /* file descriptor, output of trace (-t) is send to */
#define BUFSIZE 1024 /* size of a single buffer used with non-insert-mode */
#define IBUFSIZE 512 /* size of buffer used with insert-mode */

typedef struct box{
  struct box *next;
  int        num_args;
  char       buffer[BUFSIZE];
}box_t;

/* global variables */
int insert, trace, prompt;

int is_seperator(char c){
  char seps[]=DEFAULT_SEPERATORS; /* accepted seperators */
  int  i=0;

  while(seps[i]!=0){
    if(seps[i++]==c) return(1);
  }
  return(0);
}

proceed(FILE *fh){
  char kbdbuf[32];

  if(prompt){
    fprintf(ECHO_TO," ?... ");
    fscanf(fh,"%s",kbdbuf);
    if(kbdbuf[0]=='y') return(1);
    else return(0);
  }
  else{
    if(trace) fprintf(ECHO_TO,"\n");
    return(1);
  }
}

void err_exit(char *me, char *err_msg){
  fprintf(stderr,"%s: ",me);
  perror(err_msg);
  exit(-1);
}

main(int argc, char **argv){
  char   c, *s, *m, *end, *comm, *replstr, *xarg, *arg, **arg_vec;
  int    i, j, r, b, start_arg, arg_len, repl, stripped, arglen, repllen;
  int    num_buf_args, num_init_args, num_stdin_args, exit_status;
  int    inblk;
  pid_t  id;
  box_t  *anchor, *u, *v;
  FILE   *termfh;

  /* set defaults */
  trace=DEFAULT_TRACE;
  insert=DEFAULT_INSERT;
  prompt=DEFAULT_PROMPT;
  replstr=DEFAULT_REPLSTR;

  /* read xargs' private arguments */
  i=1; xarg=argv[1];
  while(xarg!=NULL && *xarg++=='-'){
    inblk=1;
    while(inblk&&*xarg){
      switch(*xarg++){
        case('t'):
          trace=1;
          break;
        case('p'):
          trace=1;
          prompt=1;
          break;
        case('i'):
          insert=1;
          if(*xarg!=0) replstr=xarg;
          inblk=0;
          break;
        default:
          fprintf(stderr, "Usage: %s [-tp] [-ireplstr] [command [arg..]]\n", argv[0]);
          exit(1);
          break;
      }
    }
    xarg=argv[++i];
  }

  if(argc==i){
    comm=NULL; /* no command set - use echo instead */
    start_arg=i;
    num_init_args=0; /* no initial arguments */
  }
  else{
    comm=argv[i]; /* command to execute */
    start_arg=i+1; /* first argument for the command */
    num_init_args=argc-start_arg; /* number of arguments passed to xargs */
  }

  if(prompt){
    if((termfh=fopen(ctermid(NULL), "r"))==NULL)
      err_exit(argv[0],"fopen terminal");
  }

  if(insert){
    if((arg=(char*)malloc(IBUFSIZE))==NULL)
      err_exit(argv[0],"malloc");
    if((arg_vec=(char**)malloc((num_init_args+3)*sizeof(char*)))==NULL)
      err_exit(argv[0],"malloc");
    repllen=strlen(replstr);
    arg_vec[0]=comm; /* set name of command in arguments list */
    b=0;
    while((c=getc(stdin))!=EOF){
      if(b>0 && is_seperator(c)){
        arg[b]=0; arglen=b;
        b=0;
        if(comm==NULL){
          /* Although the `echo' command isn't executed here,
             it is printed if `-t' is used. */
          if(trace) fprintf(ECHO_TO,"echo %s",arg);
          if(proceed(termfh)) printf("%s\n",arg);
        }
        else{
          if(trace) fprintf(ECHO_TO,"%s",comm);
          repl=0;
          for(i=start_arg; i<argc; i++){
            j=i-start_arg;
            r=0; s=argv[i];
            while((s=strstr(s,replstr))!=NULL){
              r++; s+=repllen;
            }
            if(r==0) arg_vec[j+1]=argv[i];
            else{
              repl=1;
              stripped=strlen(argv[i])-(r*repllen);
              if((m=arg_vec[j+1]=(char*)malloc(stripped+r*arglen+1))==NULL)
                err_exit(argv[0],"malloc");
              *m=0; s=argv[i];
              while((end=strstr(s,replstr))!=NULL){
                while(s<end) *m++=*s++;
                strcpy(m,arg);
                m+=arglen;
                s+=repllen;
              }
              strcpy(m,s);
            }
            if(trace) fprintf(ECHO_TO," %s",arg_vec[j+1]);
          }
          if(repl) arg_vec[num_init_args+1]=NULL;
          else{
            arg_vec[num_init_args+1]=arg;
            arg_vec[num_init_args+2]=NULL;
            if(trace) fprintf(ECHO_TO," %s", arg);
          }
          if(proceed(termfh)){
            if((id=fork())==(-1)) err_exit(argv[0],"fork");
            else{
              if(id==0){ /* child process */
                execvp(comm,arg_vec); /* execute command with arguments list */
                /* it's an error to step beyond this line */
                err_exit(argv[0],"execvp");
              }
              else{
                wait(&exit_status); /* parent process waits for the child */
                if(WIFEXITED(exit_status) && WEXITSTATUS(exit_status)==(unsigned char)(-1))
                  exit(-1);
              }
            }
          }
        }
      }
      else {
        if(b>0 || !is_seperator(c)) {
          /* char isn't a separator, so write new char to buffer */
          if(b<IBUFSIZE-1) arg[b++]=c;
          else {
            fprintf(stderr,"%s: buffer overflow\n",argv[0]); exit(1);
          }
        }
      }
    }
  }
  else{ /* non-insert-mode (default-mode) */
    if(trace){ /* trace-mode */
      if(comm) fprintf(ECHO_TO,"%s",comm);
      else fprintf(ECHO_TO,"echo");
    }
    b=BUFSIZE; anchor=NULL;
    while((c=getc(stdin))!=EOF){
      if(b==BUFSIZE){ /* buffer is full */
        if((u=(box_t*)malloc(sizeof(box_t)))==NULL)
          err_exit(argv[0],"malloc");
        if(anchor){
          j=0; while(v->buffer[BUFSIZE-j-1]!=0){
            if(j++==BUFSIZE){
              fprintf(stderr,"%s: buffer overflow\n",argv[0]); exit(1);
            }
          }
          for(i=0; i<j; i++) u->buffer[i]=v->buffer[BUFSIZE-j+i];
          v->num_args=num_buf_args;
          v->next=u;
          v=u;
          num_stdin_args+=num_buf_args;
        }
        else{
          anchor=v=u;
          num_stdin_args=0;
        }
        b=0;
        num_buf_args=0;
      }
      if(is_seperator(c)){
        if((b>0)&&(v->buffer[b-1]!=0)){
          v->buffer[b++]=0;
          num_buf_args++;
        }
      }
      else v->buffer[b++]=c;
    }
    if(anchor){ /* there are arguments read from stdin */
      v->num_args=num_buf_args;
      v->next=NULL;
      num_stdin_args+=num_buf_args;
      if((arg_vec=(char**)malloc((num_init_args+num_stdin_args+2)*sizeof(char*)))==NULL)
        err_exit(argv[0],"malloc");
      arg_vec[0]=comm;
      for(j=1;j<argc-start_arg+1;j++){ /* initial aruments */
        arg_vec[j]=argv[j-1+start_arg];
        if(trace) fprintf(ECHO_TO," %s",arg_vec[j]);
      }
      v=anchor;
      while(v){ /* arguments read from stdin */
        s=v->buffer;
        for(i=0;i<(v->num_args);i++){
          if(trace) fprintf(ECHO_TO," %s",s);
          arg_vec[j++]=s++;
          while(*(s-1)!=0) s++;
        }
        v=v->next;
      }
      if(proceed(termfh)){
        arg_vec[j]=NULL;
        if(comm){ /* execute command */
          execvp(comm,arg_vec);
          err_exit(argv[0],"execvp");
        }
        else{
          for(i=1;i<j;i++) printf("%s ",arg_vec[i]);
          printf("\n");
        }
      }
    }
  }
  if(prompt) fclose(termfh);
} 
